"""
This is the module that contains the main transformation from ufl to loopy
(with pdelab as the hardcoded generation target)
"""

from __future__ import absolute_import

from dune.perftool import Restriction
from dune.perftool.ufl.modified_terminals import ModifiedTerminalTracker
from dune.perftool.pymbolic.uflmapper import UFL2PymbolicMapper
from dune.perftool.pdelab.geometry import GeometryMapper
from dune.perftool.generation import (domain,
                                      get_temporary_name,
                                      global_context,
                                      globalarg,
                                      iname,
                                      instruction,
                                      symbol,
                                      temporary_variable,
                                      valuearg,
                                      )

from dune.perftool.pdelab.basis import (lfs_iname,
                                        name_lfs,
                                        name_lfs_bound,
                                        traverse_lfs_tree,
                                        )
from dune.perftool.pdelab.quadrature import quadrature_iname
from pymbolic.primitives import Subscript, Variable


@iname
def index_sum_iname(i):
    from dune.perftool.pdelab import name_index
    return name_index(i)


class UFL2LoopyVisitor(ModifiedTerminalTracker, UFL2PymbolicMapper, GeometryMapper):
    def __init__(self, measure):
        self.iname_stack = [quadrature_iname()]
        self.measure = measure
        super(UFL2LoopyVisitor, self).__init__()

    def _assign(self, o):
        # In some corner cases we do not even need a temporary variable
        if isinstance(o, int):
            return o

        # Assign a name to the temporary variable we want our result in
        temp = get_temporary_name()

        # Now we assign this expression to a new temporary variable
        insn_id = instruction(assignee=Variable(temp),
                              expression=o,
                              forced_iname_deps=frozenset({i for i in self.iname_stack}),
                              forced_iname_deps_is_final=True,
                              )

        # Actually, if we have a cache hit, we need to change our temporary
        from dune.perftool.generation import retrieve_cache_items
        temp = filter(lambda i: i.id == insn_id, retrieve_cache_items("instruction"))[0].assignee_name

        # Now that we know its exact name, declare the temporary
        temporary_variable(temp)

        return Variable(temp)

    def __call__(self, o):
        # First we do the tree traversal to get a pymbolic expression representing this expression
        ret = self.call(o)

        if isinstance(ret, Variable):
            return ret
        else:
            return self._assign(ret)

    def argument(self, o):
        # Have the issued instruction depend on the iname for this localfunction space
        from dune.perftool.pdelab.basis import lfs_iname
        self.iname_stack.append(lfs_iname(o.element(), argnumber=o.number()))

        # Correct the restriction on boundary integrals
        restriction = self.restriction
        if self.measure == 'exterior_facet':
            restriction = Restriction.NEGATIVE

        if self.grad:
            from dune.perftool.pdelab.argument import name_testfunction_gradient
            return Subscript(Variable(name_testfunction_gradient(o, restriction)), (Variable(lfs_iname(o.element(), argnumber=o.number())),))
        else:
            from dune.perftool.pdelab.argument import name_testfunction
            return Subscript(Variable(name_testfunction(o, restriction)), (Variable(lfs_iname(o.element(), argnumber=o.number())),))

    def coefficient(self, o):
        # If this is a trialfunction, we do something entirely different
        if o.count() == 0:
            # Correct the restriction on boundary integrals
            restriction = self.restriction
            if self.measure == 'exterior_facet':
                restriction = Restriction.NEGATIVE

            if self.grad:
                from dune.perftool.pdelab.argument import name_trialfunction_gradient
                return Variable(name_trialfunction_gradient(o, restriction))
            else:
                from dune.perftool.pdelab.argument import name_trialfunction
                return Variable(name_trialfunction(o, restriction))

        # so this is a parameter function
        else:
            # We expect all coefficients to be of type Expression!
            from dune.perftool.ufl.execution import Expression
            assert isinstance(o, Expression)

            # Determine the name of the parameter function
            from dune.perftool.generation import get_global_context_value
            name = get_global_context_value("namedata")[id(o)]

            # Trigger the generation of code for this thing in the parameter class
            from dune.perftool.pdelab.parameter import (cell_parameter_function,
                                                        intersection_parameter_function,
                                                        )
            if o.on_intersection:
                intersection_parameter_function(name, o)
            else:
                cell_parameter_function(name, o, self.restriction)

            # And return a symbol
            return Variable(name)

    def index_sum(self, o):
        from loopy import Reduction
        from dune.perftool.ufl.shape import determine_shape

        inames = ()
        # Define an iname for each of the indices in the multiindex
        for i in o.ufl_operands[1].indices():
            inames = inames + (index_sum_iname(i),)
            shape = determine_shape(o.ufl_operands[0], i)
            from dune.perftool.pdelab import name_index
            domain(name_index(i), shape)

        # Recurse to get the summation expression
        term = self.call(o.ufl_operands[0])
        ret = self._assign(Reduction("sum", inames, term))

        return ret


def transform_accumulation_term(term, measure, subdomain_id):
    # Initialize the local function spaces that we might need for this term
    # We therefore gather a list of modified trial functions too.
    from dune.perftool.ufl.modified_terminals import extract_modified_arguments
    test_ma = extract_modified_arguments(term)
    trial_ma = extract_modified_arguments(term, trialfunction=True, testfunction=False)

    # All test and trial functions on boundary integrals are technically restricted
    import itertools
    for ma in itertools.chain(test_ma, trial_ma):
        if measure == 'exterior_facet':
            ma.restriction = Restriction.NEGATIVE

        traverse_lfs_tree(ma)

    pymbolic_expr = UFL2LoopyVisitor(measure)(term)

    # It may happen that an entire accumulation term vanishes!
    if pymbolic_expr == 0:
        return

    # The data that is used to collect the arguments for the accumulate function
    accumargs = [None] * (2 * len(test_ma))
    residual_shape = [None] * len(test_ma)
    arg_restr = [None] * len(test_ma)
    accum_inames = [None] * len(test_ma)

    for ma in test_ma:
        count = ma.argexpr.number()
        lfs = name_lfs(ma.argexpr.element())
        lfsi = lfs_iname(ma.argexpr.element(), argnumber=count)

        accumargs[2 * count] = lfs
        accumargs[2 * count + 1] = lfsi

        accum_inames[count] = lfsi
        arg_restr[count] = ma.restriction
        residual_shape[count] = name_lfs_bound(lfs)

    from dune.perftool.pdelab.argument import name_accumulation_variable
    accumvar = name_accumulation_variable(arg_restr)

    # The residual/the jacobian should be represented through a loopy global argument
    # TODO this seems still a bit hacky, esp. w.r.t. systems
    globalarg(accumvar, shape=tuple(residual_shape))

    from dune.perftool.pdelab.quadrature import name_factor
    factor = name_factor()

    # Generate the code snippet for this accumulation instruction
    code = "{}.accumulate({}, {}*{});".format(accumvar,
                                              ", ".join(accumargs),
                                              pymbolic_expr.name,
                                              factor,
                                              )

    # Maybe wrap this instruction into a condiditional. This mostly happens with mixed boundary conditions
    if subdomain_id not in ['everywhere', 'otherwise']:
        # We need to reconstruct the subdomain_data parameter of the measure
        # I am *totally* confused as to why this information is not at hand anyway,
        # but conversation with Martin pointed me to dolfin.fem.assembly where this
        # is done in preprocessing with the limitation of only one possible type of
        # modified measure per integral type.

        # Get the original form and inspect the present measures
        from dune.perftool.generation import get_global_context_value
        original_form = get_global_context_value("formdata").original_form
        sd = original_form.subdomain_data()
        assert len(sd) == 1
        subdomains, = list(sd.values())
        domain, = list(sd.keys())
        for k in list(subdomains.keys()):
            if subdomains[k] is None:
                del subdomains[k]

        # Finally extract the original subdomain_data (which needs to be present!)
        assert measure in subdomains
        subdomain_data = subdomains[measure]

        # Determine the name of the parameter function
        name = get_global_context_value("namedata")[id(subdomain_data)]

        # Trigger the generation of code for this thing in the parameter class
        from dune.perftool.pdelab.parameter import intersection_parameter_function
        intersection_parameter_function(name, subdomain_data, t='int')

        code = "if ({} == {})\n  {}".format(name, subdomain_id, code)

    # Finally, issue the instruction
    instruction(code=code,
                assignees=frozenset({accumvar}),
                read_variables=frozenset({accumvar, factor, pymbolic_expr.name}),
                forced_iname_deps=frozenset(accum_inames).union(frozenset({quadrature_iname()})),
                forced_iname_deps_is_final=True,
                )
